package com.winhc.task.framework.es;

import com.alibaba.fastjson.JSONObject;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.winhc.task.util.QueryBuilderUtils;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchScrollRequest;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.WrapperQueryBuilder;
import org.elasticsearch.search.Scroll;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

/**
 * @author: XuJiakai
 * 2020/11/16 19:23
 */
@Slf4j
public class EsFastScan {
    private final Scroll scroll = new Scroll(TimeValue.timeValueMinutes(8L));

    private RestHighLevelClient client;
    private Consumer<SearchHit[]> func;
    private String index;
    private String type;
    private String dsl;
    private int batchSize = 10000;
    private int thread_num = 1;
    private int scanSize = 0;
    private String routing = null;
    private long totalHits = 0;

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type) {
        this(client, func, index, type, null, null);
    }

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type, String dsl, String routing) {
        this.client = client;
        this.func = func;
        this.index = index;
        this.type = type;
        this.dsl = dsl;
        this.routing = routing;
    }

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type, String dsl, int batchSize, int thread_num) {
        this.client = client;
        this.func = func;
        this.index = index;
        this.type = type;
        this.dsl = dsl;
        this.batchSize = batchSize;
        this.thread_num = thread_num;
    }


    public EsFastScan batchSize(int batchSize) {
        this.batchSize = batchSize;
        return this;
    }


    @SneakyThrows
    public void scan() {
        int poolSize = 6;
        ArrayBlockingQueue<Runnable> objects = new ArrayBlockingQueue<>(poolSize * 2);
        ThreadPoolExecutor executorService = new ThreadPoolExecutor(
                poolSize, poolSize,
                0L, TimeUnit.MILLISECONDS,
                objects,
                new ThreadFactoryBuilder().setNameFormat("ScanEs-pool").build(),
                (r, executor) -> {
                    try {
                        executor.getQueue().put(r);
                    } catch (InterruptedException e) {
                        throw new RejectedExecutionException("interrupted", e);
                    }
                }
        );


        SearchRequest searchRequest = new SearchRequest(index).types(type).scroll(scroll);
        if (routing != null) {
            searchRequest.routing(routing);
        }

        SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder().size(batchSize);
        if (StringUtils.isNotEmpty(dsl)) {
            WrapperQueryBuilder query = QueryBuilderUtils.getQuery(JSONObject.parseObject(dsl));
            searchSourceBuilder.query(query);
        }

        searchRequest.source(searchSourceBuilder);
        SearchResponse searchResponse = client.search(searchRequest);

        totalHits = searchResponse.getHits().getTotalHits();
        log.info("es scan start shard total hits: {}", totalHits);
        int n = searchResponse.getHits().getHits().length;
        if (n != 0) {
            executorService.submit(() -> {
                scanSize += n;
                log.info("scan size : {} , shard id : {} , shard size: {}, 执行一批：{}", scanSize, routing, totalHits, n);
                func.accept(searchResponse.getHits().getHits());
            });
        } else {
            log.warn("数据为空！");
        }

        do {
            int num = scanData(executorService, searchResponse.getScrollId());
            if (num == 0) {
                break;
            }
        } while (true);

        executorService.shutdown();
        while (!executorService.awaitTermination(1, TimeUnit.MINUTES)) {
            System.out.println("线程池没有关闭");
        }

        log.info("es scan is successful，total hits: {}", totalHits);
    }

    @SneakyThrows
    private int scanData(ThreadPoolExecutor executorService, String scrollId) {
        SearchScrollRequest scroll = new SearchScrollRequest(scrollId).scroll(this.scroll);
        SearchResponse searchResponse = client.searchScroll(scroll);
        int num = searchResponse.getHits().getHits().length;
        if (num == 0) {
            return num;
        }
        executorService.submit(() -> {
            scanSize += num;
            log.info("scan size : {} , shard id : {} , shard size: {}, 执行一批：{}", scanSize, routing, totalHits, num);
            func.accept(searchResponse.getHits().getHits());
        });
        return num;
    }
}
