package com.winhc.task.job;

import cn.hutool.cache.CacheUtil;
import cn.hutool.cache.impl.TimedCache;
import com.aliyuncs.dataworks_public.model.v20180601.CreateManualDagResponse;
import com.winhc.task.bean.*;
import com.winhc.task.service.TouchService;
import com.winhc.task.util.DateUtils;
import com.winhc.task.util.SchemaInit;
import com.winhc.task.util.SmsUtils;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author π
 * @Description:
 * @date 2021/12/14 18:23
 */
@Slf4j
@Service
public class MultipleSynHoloJob {

    @Autowired
    TouchService touchService;

    @SneakyThrows
    public void start(List<String> tns, JobArgs jobArgs) {
        List<DataWorksFlowJob> jobs = SchemaInit.getJobs(tns, jobArgs);
        String bizDate = DateUtils.getYesterday();
        Set<TaskInfo> failureTask = run(bizDate, jobs);
        if (failureTask.size() > 0) {
            Set<String> fSet = failureTask.stream().map(TaskInfo::getKey).collect(Collectors.toSet());
            log.error("job syn es error list : {}", fSet);
            System.exit(-1);
        }
        log.info("end MultipleSynHoloJob !");
    }

    @SneakyThrows
    public void start(List<String> tns, SynKafKaJobArgs jobArgs) {
        List<DataWorksFlowJob> jobs = SchemaInit.getJobs2(tns, jobArgs);
        Set<TaskInfo> failureTask = run(DateUtils.getYesterday(), jobs);
        if (failureTask.size() > 0) {
            Set<String> fSet = failureTask.stream().map(TaskInfo::getKey).collect(Collectors.toSet());
            log.error("job syn es error list : {}", fSet);
            System.exit(-1);
        }
        log.info("end MultipleSynHoloJob !");
    }

    @SneakyThrows
    public Set<TaskInfo> run(String bizDate, List<DataWorksFlowJob> jobs) {
        log.info("start MultipleSynHoloJob !");
        LocalDateTime start = LocalDateTime.now();
        List<TaskInfo> collect = jobs.stream()
                .flatMap(dataWorksFlowJob -> {
                    List<DataWorksFlowTask> task = dataWorksFlowJob.getTask();
                    return task
                            .stream()
                            .map(dataWorksFlowTask -> {
                                TaskParam build = TaskParam.builder()
                                        .projectName(dataWorksFlowJob.getProject())
                                        .bizDate(bizDate)
                                        .flowName(dataWorksFlowJob.getFlow())
                                        .nodeParam(dataWorksFlowTask.toNodeParam(bizDate))
                                        .build();
                                CreateManualDagResponse r = touchService.touch(build);

                                return TaskInfo.builder().project(dataWorksFlowJob.getProject())
                                        .flow(dataWorksFlowJob.getFlow())
                                        .taskName(dataWorksFlowTask.getTaskName())
                                        .dagId(r.getReturnValue()).build();
                            });
                }).collect(Collectors.toList());

        int totalTask = collect.size();

        Set<TaskInfo> end = new HashSet<>();
        Set<TaskInfo> failureTask = new HashSet<>();
        TimedCache<TaskInfo, String> timedCache = CacheUtil.newTimedCache(300 * 1000);
        int i = 0;
        int successTask = 0;
        int failedTask = 0;
        while (true) {
            int awaitTask = 0;

            boolean flag = true;
            List<String> empty = new ArrayList<>();

            for (TaskInfo taskInfo : collect) {
                if (end.contains(taskInfo)) {
                    continue;
                }
                Map<String, TaskFlowEnum> query = touchService.query(taskInfo.getProject(), taskInfo.getDagId());
                Map<String, List<String>> status = query.entrySet().stream()
                        .map(e -> {
                            Entry<String, String> entry = new Entry<>();
                            if (e.getValue().equals(TaskFlowEnum.SUCCESS) || e.getValue().equals(TaskFlowEnum.FAILURE)) {
                                entry.setKey(e.getValue().getMsg());
                            } else {
                                entry.setKey("运行中");
                            }
                            entry.setValue(e.getKey());
                            return entry;
                        }).collect(Collectors.groupingBy(Entry::getKey, Collectors.mapping(Entry::getValue, Collectors.toList())));
                List<String> success = status.getOrDefault(TaskFlowEnum.SUCCESS.getMsg(), empty);
                List<String> failure = status.getOrDefault(TaskFlowEnum.FAILURE.getMsg(), empty);
                List<String> await = status.getOrDefault(TaskFlowEnum.RUNNING.getMsg(), empty);

                if (failure.size() != 0) {
                    failedTask++;
                    log.error("failure node:{} ", failure);
                    DingMsg error = new DingMsg("任务失败", taskInfo.getProject(), taskInfo.getFlow(), taskInfo.getTaskName(), String.join(",", failure), TaskFlowEnum.FAILURE.getMsg());
                    log.error(error.toMd());
                    failureTask.add(taskInfo);
                } else {
                    if (await.size() != 0) {
                        awaitTask++;
                    } else {
                        successTask++;
                    }
                }
                flag = flag && (await.size() == 0 || failure.size() != 0);
                if (await.size() == 0 || failure.size() != 0) {
                    end.add(taskInfo);
                } else {
                    LocalDateTime now = LocalDateTime.now();
                    Duration duration = Duration.between(start, now);
                    long l = duration.toHours();
                    if (l >= 2) {
                        if (!timedCache.containsKey(taskInfo) && i <= 6) {
                            //超两小时
                            i++;
                            DingMsg error = new DingMsg("【" + i + "】任务长时间未结束", taskInfo.getProject(), taskInfo.getFlow(), taskInfo.getTaskName(), String.join(",", failure), TaskFlowEnum.RUNNING.getMsg());
                            log.error(error.toMd());
                            timedCache.put(taskInfo, "1");
                        }
                    }
                }
            }
            log.info("\nawait task:{}\ntotal task:{}\nsuccess task:{}\nfailure task:{}\n"
                    , awaitTask
                    , totalTask
                    , successTask
                    , failedTask
            );
            if (flag) {
                break;
            }
            Thread.sleep(10000);
        }
        return failureTask;
    }
}
